/*

Copyright  2004 Cornell Research Foundation, Inc.  All rights reserved.
Author:  Prof. Martin Burtscher

Software License Terms and Conditions

1.  SOFTWARE shall mean VPC3 code or portions thereof contained in the file
    VPC3.tar.gz and described in Cornell Research Foundation, Inc. ("CRF")
    file D-3392.  SOFTWARE includes, but is not limited to, VPC3 source code,
    object code and executable code.

2.  CRF is a wholly owned subsidiary of Cornell University, is a fiduciary of
    Cornell University in intellectual property matters and holds all
    intellectual property rights in SOFTWARE.

3.  LICENSEE means the party to this Agreement and the user of SOFTWARE.  By
    using SOFTWARE, LICENSEE enters into this Agreement with CRF.

4.  SOFTWARE is made available under this Agreement to allow certain non-
    commercial research and classroom use.  CRF reserves all commercial rights
    to SOFTWARE and these rights may be licensed by CRF to third parties.

5.  LICENSEE is hereby granted permission to download, compile, execute, copy,
    and modify SOFTWARE for non-commercial research or teaching purposes
    provided that this notice accompanies all copies of SOFTWARE.  Copies of
    modified SOFTWARE may be distributed only for non-commercial research or
    teaching purposes (a) if this notice accompanies those copies, (b) if said
    copies carry prominent notices stating that SOFTWARE has been changed, and
    (c) the date of any changes are clearly identified in SOFTWARE.

6.  LICENSEE agrees that the export of SOFTWARE from the United States may
    require approval from the U.S. government and failure to obtain such
    approval will result in the immediate termination of this license and may
    result in criminal liability under U.S. laws.

7.  The work leading to the development of SOFTWARE was supported in part by
    various grants from an agency of the U.S. Government, and CRF is obligated
    to comply with U.S. OMB Circular A-124 and 37 CFR Part 401.  This license
    is subject to the applicable terms of U.S. Government regulations
    concerning Government funded inventions.

8.  CRF provides SOFTWARE on an "as is" basis.  CRF does not warrant,
    guarantee, or make any representations regarding the use or results of
    SOFTWARE with respect to its correctness, accuracy, reliability or
    performance.  The entire risk of the use and performance of SOFTWARE is
    assumed by LICENSEE.  ALL WARRANTIES INCLUDING, WITHOUT LIMITATION, ANY
    WARRANTY OF FITNESS FOR A PARTICULAR PURPOSE OR MERCHANTABILITY AND ANY
    WARRANTY OF NONINFRINGEMENT OF PATENTS, COPYRIGHTS, OR ANY OTHER
    INTELLECTUAL PROPERTY RIGHT ARE HEREBY EXCLUDED.

9.  LICENSEE understands and agrees that neither CRF nor Cornell University is
    under any obligation to provide maintenance, support or update services,
    notices of latent defects, correction of defects, or future versions for
    SOFTWARE.

10. Even if advised of the possibility of damages, under no circumstances
    shall CRF or Cornell University individually or jointly be liable to
    LICENSEE or any third party for damages of any character, including,
    without limitation, direct, indirect, incidental, consequential or
    special damages, loss of profits, loss of use, loss of goodwill,
    computer failure or malfunction.  LICENSEE agrees to indemnify and hold
    harmless CRF and Cornell University for any and all liability CRF or
    Cornell University may incur as a result of use of SOFTWARE by LICENSEE.

*/


// works only on little-endian machines

#include <stdlib.h>
#include <stdio.h>
#include <assert.h>


#define pfcmsizeblg2 17  /* at least 16, or change inlined hash function */
#define vdim (1024*64)
#define vfcmsizeblg2 19
#define vdfcmsizeblg2 17

#define BufElems 8192


typedef long long etype;

static FILE *fin, *fo1a, *fo1b, *fo2a, *fo2b;

static long pused[5];
static int pfcma0, pfcma1, pfcma2;
static int *pfcmb0, *pfcmb2;

static long vused[11];
static etype *vfcmb, *vdfcmb0, *vdfcmb2;
static int vfcma[vdim], vdfcma[vdim][3];
static etype vlnv[vdim][4];


static void Check(const void * const ptr)
{
  if (ptr == NULL) {
    printf("out of memory");
    exit(-1);
  }
}


static void Init(void)
{
  register int i, size;

  for (i = 0; i < 5; i++) pused[i] = 0;
  assert(15 < pfcmsizeblg2);
  pfcma0 = 0;
  pfcma1 = 0;
  pfcma2 = 0;
  size = (1<<pfcmsizeblg2)*2;
  pfcmb0 = malloc(sizeof(int)*size);
  Check(pfcmb0);
  for (i = 0; i < size; i++) pfcmb0[i] = i & 1;
  size = (1<<(pfcmsizeblg2+2))*2;
  pfcmb2 = malloc(sizeof(int)*size);
  Check(pfcmb2);
  for (i = 0; i < size; i++) pfcmb2[i] = i & 1;


  assert(0 == (vdim & (vdim-1)));
  for (i = 0; i < 11; i++) vused[i] = 0;
  assert(0 < vdfcmsizeblg2-2);
  for (i = 0; i < vdim; i++) {
    vdfcma[i][0] = 0;
    vdfcma[i][1] = 0;
    vdfcma[i][2] = 0;
  }
  size = (1<<vdfcmsizeblg2)*2;
  vdfcmb0 = malloc(sizeof(etype)*size);
  Check(vdfcmb0);
  for (i = 0; i < size; i++) vdfcmb0[i] = i & 1;
  size = (1<<(vdfcmsizeblg2+2))*2;
  vdfcmb2 = malloc(sizeof(etype)*size);
  Check(vdfcmb2);
  for (i = 0; i < size; i++) vdfcmb2[i] = i & 1;

  for (i = 0; i < vdim; i++) {
    vlnv[i][0] = 0;
    vlnv[i][1] = 1;
    vlnv[i][2] = 2;
    vlnv[i][3] = 3;
  }

  assert(0 < vfcmsizeblg2);
  for (i = 0; i < vdim; i++) vfcma[i] = 0;
  size = (1<<vfcmsizeblg2)*2;
  vfcmb = malloc(sizeof(etype)*size);
  Check(vfcmb);
  for (i = 0; i < size; i++) vfcmb[i] = i & 1;
}


static int PCPredictUpdate(const int pc)
{
  register int pcode, phash, *pline;
  register long hicnt;

  pcode = 4;
  hicnt = -1;

  pline = &(pfcmb0[pfcma0<<1]);
  if (pc == pline[0]) {
    pcode = 0;
    hicnt = pused[0];
  }
  if ((pused[1] > hicnt) && (pc == pline[1])) {
    pcode = 1;
    hicnt = pused[1];
  }
  pline[1] = pline[0];
  pline[0] = pc;

  pline = &(pfcmb2[pfcma2<<1]);
  if ((pused[2] > hicnt) && (pc == pline[0])) {
    pcode = 2;
    hicnt = pused[2];
  }
  if ((pused[3] > hicnt) && (pc == pline[1])) {
    pcode = 3;
    hicnt = pused[3];
  }
  pline[1] = pline[0];
  pline[0] = pc;

  phash = pc ^ (pc >> pfcmsizeblg2);
  phash &= (1<<pfcmsizeblg2)-1;
  pfcma2 = (pfcma1 << 1) ^ phash;
  pfcma1 = (((((pfcma0 << pfcmsizeblg2) | pfcma0) >> 2) & ((1<<pfcmsizeblg2)-1)) << 1) ^ phash;
  pfcma0 = phash;

  pused[pcode]++;
  return pcode;
}


static int PCDecode(const int pcode, int pc)
{
  register int phash, *pline;

  switch (pcode) {
    case 0: pc = pfcmb0[pfcma0<<1]; break;
    case 1: pc = pfcmb0[(pfcma0<<1)+1]; break;
    case 2: pc = pfcmb2[pfcma2<<1]; break;
    case 3: pc = pfcmb2[(pfcma2<<1)+1]; break;
    case 4: break;  // pc already set
    default: printf("unrecognized pcode encountered\n"); exit(-1); break;
  }

  pline = &(pfcmb0[pfcma0<<1]);
  pline[1] = pline[0];
  pline[0] = pc;

  pline = &(pfcmb2[pfcma2<<1]);
  pline[1] = pline[0];
  pline[0] = pc;

  phash = pc ^ (pc >> pfcmsizeblg2);
  phash &= (1<<pfcmsizeblg2)-1;
  pfcma2 = (pfcma1 << 1) ^ phash;
  pfcma1 = (((((pfcma0 << pfcmsizeblg2) | pfcma0) >> 2) & ((1<<pfcmsizeblg2)-1)) << 1) ^ phash;
  pfcma0 = phash;

  return pc;
}


static inline int HashVal(etype x, const int bits)
{
  register int res = 0;

  while (x > 0) {
    res ^= x;
    x = (unsigned long long)x >> bits;
  }
  return (res & ((1<<bits)-1));
}


static int ValPredictUpdate(const int pc, const etype val)
{
  register int vcode, vhash, idx;
  register etype s, *vline;
  register long hicnt;

  idx = pc & (vdim-1);
  vcode = 10;
  hicnt = -1;
  s = val - vlnv[idx][0];

  vline = &(vdfcmb0[vdfcma[idx][0]<<1]);
  if (s == vline[0]) {
    vcode = 0;
    hicnt = vused[0];
  }
  if ((vused[1] > hicnt) && (s == vline[1])) {
    vcode = 1;
    hicnt = vused[1];
  }
  vline[1] = vline[0];
  vline[0] = s;

  vline = &(vdfcmb2[vdfcma[idx][2]<<1]);
  if ((vused[2] > hicnt) && (s == vline[0])) {
    vcode = 2;
    hicnt = vused[2];
  }
  if ((vused[3] > hicnt) && (s == vline[1])) {
    vcode = 3;
    hicnt = vused[3];
  }
  vline[1] = vline[0];
  vline[0] = s;

  vhash = HashVal(s, vdfcmsizeblg2);
  vdfcma[idx][2] = (vdfcma[idx][1] << 1) ^ vhash;
  vdfcma[idx][1] = (vdfcma[idx][0] << 1) ^ vhash;
  vdfcma[idx][0] = vhash;


  if ((vused[4] > hicnt) && (val == vlnv[idx][0])) {
    vcode = 4;
    hicnt = vused[4];
  }
  if ((vused[5] > hicnt) && (val == vlnv[idx][1])) {
    vcode = 5;
    hicnt = vused[5];
  }
  if ((vused[6] > hicnt) && (val == vlnv[idx][2])) {
    vcode = 6;
    hicnt = vused[6];
  }
  if ((vused[7] > hicnt) && (val == vlnv[idx][3])) {
    vcode = 7;
    hicnt = vused[7];
  }
  vlnv[idx][3] = vlnv[idx][2];
  vlnv[idx][2] = vlnv[idx][1];
  vlnv[idx][1] = vlnv[idx][0];
  vlnv[idx][0] = val;


  vline = &(vfcmb[vfcma[idx]<<1]);
  if ((vused[8] > hicnt) && (val == vline[0])) {
    vcode = 8;
    hicnt = vused[8];
  }
  if ((vused[9] > hicnt) && (val == vline[1])) {
    vcode = 9;
    hicnt = vused[9];
  }
  vline[1] = vline[0];
  vline[0] = val;
  vfcma[idx] = HashVal(val, vfcmsizeblg2);


  vused[vcode]++;
  return vcode;
}


static etype ValDecode(const int vcode, const int pc, etype val)
{
  register int vhash, idx;
  register etype s, *vline;

  idx = pc & (vdim-1);
  switch (vcode) {
    case 0: val = vdfcmb0[vdfcma[idx][0]<<1]+vlnv[idx][0]; break;
    case 1: val = vdfcmb0[(vdfcma[idx][0]<<1)+1]+vlnv[idx][0]; break;
    case 2: val = vdfcmb2[vdfcma[idx][2]<<1]+vlnv[idx][0]; break;
    case 3: val = vdfcmb2[(vdfcma[idx][2]<<1)+1]+vlnv[idx][0]; break;
    case 4: val = vlnv[idx][0]; break;
    case 5: val = vlnv[idx][1]; break;
    case 6: val = vlnv[idx][2]; break;
    case 7: val = vlnv[idx][3]; break;
    case 8: val = vfcmb[vfcma[idx]<<1]; break;
    case 9: val = vfcmb[(vfcma[idx]<<1)+1]; break;
    case 10: break;  // val already set
    default: printf("unrecognized vcode encountered\n"); exit(-1); break;
  }

  s = val - vlnv[idx][0];
  vline = &(vdfcmb0[vdfcma[idx][0]<<1]);
  vline[1] = vline[0];
  vline[0] = s;
  vline = &(vdfcmb2[vdfcma[idx][2]<<1]);
  vline[1] = vline[0];
  vline[0] = s;
  vhash = HashVal(s, vdfcmsizeblg2);
  vdfcma[idx][2] = (vdfcma[idx][1] << 1) ^ vhash;
  vdfcma[idx][1] = (vdfcma[idx][0] << 1) ^ vhash;
  vdfcma[idx][0] = vhash;

  vlnv[idx][3] = vlnv[idx][2];
  vlnv[idx][2] = vlnv[idx][1];
  vlnv[idx][1] = vlnv[idx][0];
  vlnv[idx][0] = val;

  vline = &(vfcmb[vfcma[idx]<<1]);
  vline[1] = vline[0];
  vline[0] = val;
  vfcma[idx] = HashVal(val, vfcmsizeblg2);

  return val;
}


static void Commit(const void * const buf, const int elemsize, const int count, FILE * const f, const char * const str)
{
  if (fwrite(buf, elemsize, count, f) != count) {
    printf("error writing to %s file\n", str);
    exit(-1);
  }
}


static void Compress(void)
{
  register int pc, pcode, vcode;
  register int cur1a, cur1b, cur2a, cur2b, cur, tot;
  register etype val;
  char header[4];
  char buf1a[BufElems];
  char buf2a[BufElems];
  int buf1b[BufElems];
  etype buf2b[BufElems];
  char bufin[BufElems*(sizeof(int)+sizeof(etype))];

  assert(fread(header, sizeof(int), 1, fin) == 1);  // read header
  cur1a = 0; cur1b = 0; cur2a = 0; cur2b = 0; cur = 0;
  tot = fread(bufin, 1, (sizeof(int)+sizeof(etype))*BufElems, fin);
  do {
    while (cur < tot) {
      pc = *((int*)&(bufin[cur]));
      cur += sizeof(int);
      val = *((int*)&(bufin[cur+4]));  // to prevent alignment error
      val = (val << 32) + *((unsigned int*)&(bufin[cur]));
      cur += sizeof(etype);

      pcode = PCPredictUpdate(pc);
      vcode = ValPredictUpdate(pc, val);

      buf1a[cur1a] = pcode;
      cur1a++;
      if (cur1a == BufElems) {
        Commit(buf1a, sizeof(char), BufElems, fo1a, "pc output (a)");
        cur1a = 0;
      }
      if (pcode == 4) {
        buf1b[cur1b] = pc;
        cur1b++;
        if (cur1b == BufElems) {
          Commit(buf1b, sizeof(int), BufElems, fo1b, "pc output (b)");
          cur1b = 0;
        }
      }
      buf2a[cur2a] = vcode;
      cur2a++;
      if (cur2a == BufElems) {
        Commit(buf2a, sizeof(char), BufElems, fo2a, "val output (a)");
        cur2a = 0;
      }
      if (vcode == 10) {
        buf2b[cur2b] = val;
        cur2b++;
        if (cur2b == BufElems) {
          Commit(buf2b, sizeof(etype), BufElems, fo2b, "val output (b)");
          cur2b = 0;
        }
      }
    }
    cur = 0;
    tot = fread(bufin, 1, (sizeof(int)+sizeof(etype))*BufElems, fin);
  } while (tot > 0);
  Commit(buf1a, sizeof(char), cur1a, fo1a, "pc output (a)");
  Commit(buf2a, sizeof(char), cur2a, fo2a, "val output (a)");
  Commit(buf1b, sizeof(int), cur1b, fo1b, "pc output (b)");
  Commit(buf2b, sizeof(etype), cur2b, fo2b, "val output (b)");
}


static void Decompress(void)
{
  register int pc, pcode, vcode, tot1a, cur1a, tot1b, cur1b, tot2a, cur2a, tot2b, cur2b, cur;
  register etype val;
  char header[4] = {'v', 'p', 'c', '3'};
  signed char buf1a[BufElems];
  signed char buf2a[BufElems];
  char bufin[BufElems+sizeof(etype)+sizeof(int)];
  int buf1b[BufElems];
  etype buf2b[BufElems];

  assert(fwrite(header, sizeof(int), 1, fin) == 1);  // write header
  tot1a = fread(buf1a, sizeof(char), BufElems, fo1a);
  cur1a = 0; tot1b = 0; cur1b = 0; tot2a = 0; cur2a = 0; tot2b = 0; cur2b = 0; cur = 0;
  do {
    while (cur1a < tot1a) {
      pcode = buf1a[cur1a];
      cur1a++;
      if (cur2a == tot2a) {
        tot2a = fread(buf2a, sizeof(char), BufElems, fo2a);
        cur2a = 0;  
        if (0 == tot2a) {  // not sure this is correct
          printf("pc and value file do not contain the same number of elements\n");
          exit(-1);
        }
      }
      vcode = buf2a[cur2a];
      cur2a++;
      if (pcode == 4) {
        if (cur1b == tot1b) {
          tot1b = fread(buf1b, sizeof(int), BufElems, fo1b);
          cur1b = 0;  
          if (0 == tot1b) {
            printf("error reading from pc input (b) file\n");
            exit(-1);
          }
        }
        pc = buf1b[cur1b];
        cur1b++;
      }
      if (vcode == 10) {
        if (cur2b == tot2b) {
          tot2b = fread(buf2b, sizeof(etype), BufElems, fo2b);
          cur2b = 0;
          if (0 == tot2b) {
            printf("error reading from val input (b) file\n");
            exit(-1);
          }
        }
        val = buf2b[cur2b];
        cur2b++;
      }

      pc = PCDecode(pcode, pc);
      val = ValDecode(vcode, pc, val);

      *((int*)&(bufin[cur])) = pc;
      cur += sizeof(int);
      *((int*)&(bufin[cur])) = val;
      *((int*)&(bufin[cur+4])) = val >> 32;
      cur += sizeof(etype);
      if (cur >= BufElems) {
        Commit(bufin, 1, cur, fin, "output");
        cur = 0;
      }
    }
    tot1a = fread(buf1a, sizeof(char), BufElems, fo1a);
    cur1a = 0;
  } while (tot1a > 0);
  Commit(bufin, 1, cur, fin, "output");
}


static FILE *Open(const char * const fname, const char * const mode)
{
  FILE *f;

  f = fopen(fname, mode);
  if (f == NULL) {
    printf("could not open file %s\n", fname);
    exit(-1);
  }
  return f;
}


int main(int argc, char *argv[])
{
  char fname[256];

  assert(sizeof(int) == 4);
  assert(sizeof(etype) == 8);
  assert(0 == (BufElems & (BufElems-1)));

  if (argc == 2) {
    sprintf(fname, "%s", argv[1]);
    fin = Open(fname, "r+b");
    sprintf(fname, "%s.stream1", argv[1]);
    fo1a = Open(fname, "w+b");
    sprintf(fname, "%s.stream2", argv[1]);
    fo1b = Open(fname, "w+b");
    sprintf(fname, "%s.stream3", argv[1]);
    fo2a = Open(fname, "w+b");
    sprintf(fname, "%s.stream4", argv[1]);
    fo2b = Open(fname, "w+b");

    Init();
    Compress();

    fclose(fin);
    fclose(fo1a);
    fclose(fo1b);
    fclose(fo2a);
    fclose(fo2b);
  } else if ((argc == 3) && (argv[1][0] == 'd') && (argv[1][1] == 0)) {
    sprintf(fname, "%s.decom", argv[2]);
    fin = Open(fname, "w+b");
    sprintf(fname, "%s.stream1", argv[2]);
    fo1a = Open(fname, "r+b");
    sprintf(fname, "%s.stream2", argv[2]);
    fo1b = Open(fname, "r+b");
    sprintf(fname, "%s.stream3", argv[2]);
    fo2a = Open(fname, "r+b");
    sprintf(fname, "%s.stream4", argv[2]);
    fo2b = Open(fname, "r+b");

    Init();
    Decompress();

    fclose(fin);
    fclose(fo1a);
    fclose(fo1b);
    fclose(fo2a);
    fclose(fo2b);
  } else {
    printf("usage: %s [d] tracename\n", argv[0]);
  }

  return 0;
}

